
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"

#include "include_file.h"
#include "pass_4.h"
#include "parse.h"
#include "stack.h"


extern struct section_def *sections_first, *sections_last, *sec_tmp, *sec_next;
extern struct incbin_file_data *incbin_file_data_first, *ifd_tmp;
extern struct export_def *export_first, *export_last, *export_tmp;
extern struct stack *stacks_first, *stacks_tmp, *stacks_last, *stacks_header_first, *stacks_header_last;
extern struct label_def *label_tmp, *label_last, *labels;
extern struct gba_def *defines, *tmp_def, *next_def;
extern struct slot slots[256];
extern FILE *file_out_ptr;
extern unsigned char *rom_banks, *rom_banks_usage_table;
extern char gba_tmp_name[32], tmp[4096], name[32], final_name[MAX_NAME_LENGTH];
extern int rombanks, ind, inz, output_format, compact_mode, test_mode;

#ifdef GB
extern char licenseecodenew_c1, licenseecodenew_c2;
extern int computechecksum_defined, computecomplementcheck_defined;
extern int cartridgetype;
extern int name_defined, computechecksum_defined, computecomplementcheck_defined;
extern int licenseecodeold_defined, licenseecodenew_defined;
extern int licenseecodeold, romgbc;
#endif

extern int romtype, rambanks, emptyfill, max_address;
extern int rambanks_defined, verbose_mode;
extern int section_status;
extern int banksize, banksize_defined;
extern int slots_amount;
extern int error_level;
extern int *banks, *bankaddress, rombankmap_defined;
extern int latest_stack, stacks_outside, stacks_inside;
extern int bankheader_status;
extern int symbol_mode;

#ifndef GB
extern int header, header_defined;
extern int footer, footer_defined;
int final_size, binary_start;
#endif

#ifdef W65816
extern int lorom_defined, hirom_defined, slowrom_defined, fastrom_defined, snes_mode;
#endif

#ifdef Z80
extern int computesmschecksum_defined, smstag_defined;
#endif

struct label_def *unknown_labels = NULL, *unknown_labels_last = NULL, *tul, *ltmp;
struct label_def *unknown_header_labels = NULL, *unknown_header_labels_last = NULL;

char lb[16], symb[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
unsigned int pc_bank = 0, pc_full = 0, rom_bank, mem_insert_overwrite, symb_size, slot, pc_slot, pc_slot_max;
int x, q, ov;
int filename_id, line_number;


#define WRITEOUT_OV fprintf(final_ptr, "%c%c%c%c", (ov>>24) & 0xFF, (ov>>16) & 0xFF, (ov>>8) & 0xFF, ov & 0xFF);



int compact_mode_next_name(void) {

  static int c1 = 0, c2 = 0, p = 1;


  if (p == 1) {
    if (c1 == symb_size) {
      p++;
      lb[p + 1] = 0;
      c1 = 0;
      c2 = 0;
      lb[p - 1] = symb[c1];
      lb[p] = symb[c2];
    }
    else {
      lb[p] = symb[c1];
      c1++;
    }
  }
  else if (p == 2) {
    if (c1 == symb_size) {
      c1 = 0;
      c2++;
      lb[p - 1] = symb[c1];
      lb[p] = symb[c2];
    }
    else {
      lb[p - 1] = symb[c1];
      c1++;
    }
    if (c2 == symb_size) {
      p++;
      lb[p + 1] = 0;
      c1 = 0;
      c2 = 0;
      lb[p - 1] = symb[c1];
      lb[p] = symb[c2];
    }
  }

  ltmp = labels;
  while (ltmp != NULL) {
    if (strcmp(ltmp->label, lb) == 0) {
      return FAILED;
    }
    ltmp = ltmp->next;
  }

  return SUCCEEDED;
}


int new_unknown_reference(int type) {

  struct label_def *t;


  t = malloc(sizeof(struct label_def));
  if (t == NULL) {
    if (error_level == ERROR_LEVEL_MAX)
      fprintf(stderr, "%s:%d: NEW_UNKNOWN_REFERENCE: Out of memory.\n", get_file_name(filename_id), line_number);
    else
      fprintf(stderr, "NEW_UNKNOWN_REFERENCE: Out of memory.\n");

    return FAILED;
  }

  if (error_level == ERROR_LEVEL_MAX) {
    t->filename_id = filename_id;
    t->linenumber = line_number;
  }

  strcpy(t->label, tmp);
  t->next = NULL;
  t->type = type;
  t->section_status = section_status;
  if (section_status == ON) {
    t->section_id = sec_tmp->id;
    t->address = sec_tmp->i; /* relative address, to the beginning of the section */
  }
  else
    t->address = pc_bank; /* bank address, in ROM memory */

  t->bank = rom_bank;
  t->slot = slot;

  /* outside bank header section */
  if (bankheader_status == OFF) {
    if (unknown_labels_last == NULL) {
      unknown_labels = t;
      unknown_labels_last = t;
    }
    else {
      unknown_labels_last->next = t;
      unknown_labels_last = t;
    }
  }
  /* bank header section */
  else {
    if (t->label[0] == '_') {
      if (error_level == ERROR_LEVEL_MAX)
	fprintf(stderr, "%s:%d: NEW_UNKNOWN_REFERENCE: Referring to a local label (\"%s\") from inside a bank header section is not allowed.\n", get_file_name(filename_id), line_number, t->label);
      else
	fprintf(stderr, "NEW_UNKNOWN_REFERENCE: Referring to a local label (\"%s\") from inside a bank header section is not allowed.\n", t->label);
      return FAILED;
    }
    if (unknown_header_labels_last == NULL) {
      unknown_header_labels = t;
      unknown_header_labels_last = t;
    }
    else {
      unknown_header_labels_last->next = t;
      unknown_header_labels_last = t;
    }
  }

  return SUCCEEDED;
}


int pass_4(void) {

  FILE *final_ptr;
  char *t, c;
  int i, o = 0, z, y;
#ifndef GB
  int header_size = 0, footer_size = 0;
#endif
#ifdef W65816
  int base = 0x00;
#endif


  section_status = OFF;
  bankheader_status = OFF;
  mem_insert_overwrite = OFF;

  if (verbose_mode == ON)
    printf("Internal pass 2...\n");

  if ((file_out_ptr = fopen(gba_tmp_name, "rb")) == NULL) {
    fprintf(stderr, "INTERNAL_PASS_2: Error opening file \"%s\".\n", gba_tmp_name);
    return FAILED;
  }

  while (fread(&c, 1, 1, file_out_ptr) != 0) {

    switch (c) {

    case 'E':
      continue;

    case 'f':
      fscanf(file_out_ptr, "%d ", &filename_id);
      continue;

    case 'k':
      fscanf(file_out_ptr, "%d ", &line_number);
      continue;

    /* SECTION */

    case 'A':
    case 'S':
      if (c == 'A')
	fscanf(file_out_ptr, "%d %d ", &x, &ind);
      else
	fscanf(file_out_ptr, "%d ", &x);

      sec_tmp = sections_first;
      while (sec_tmp != NULL) {
	if (sec_tmp->id == x)
	  break;
	sec_tmp = sec_tmp->next;
      }

      if (c == 'A')
	sec_tmp->address = ind;

      ind = 0;
      if (sec_tmp->maxsize_status == ON) {
	if (sec_tmp->maxsize < sec_tmp->size) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s: INTERNAL_PASS_2: Section \"%s\" doesn't fit into the specified %d bytes. Enlarging to %d bytes.\n",
		    get_file_name(filename_id), sec_tmp->name, sec_tmp->maxsize, sec_tmp->size);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Section \"%s\" doesn't fit into the specified %d bytes. Enlarging to %d bytes.\n",
		    sec_tmp->name, sec_tmp->maxsize, sec_tmp->size);
	}
	else if (sec_tmp->size < sec_tmp->maxsize) {
	  sec_tmp->size = sec_tmp->maxsize;
	  ind = 1;
	}
      }

      sec_tmp->data = malloc(sizeof(unsigned char) * sec_tmp->size);

      if (sec_tmp->data == NULL) {
	if (error_level == ERROR_LEVEL_MAX)
	  fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Out of memory when trying to allocate room for section \"%s\".\n",
		  get_file_name(filename_id), line_number, sec_tmp->name);
	else
	  fprintf(stderr, "INTERNAL_PASS_2: Out of memory when trying to allocate room for section \"%s\".\n", sec_tmp->name);
	return FAILED;
      }

      /* fill the padded area with _emptyfill_ */
      if (ind == 1)
	memset(sec_tmp->data, emptyfill, sec_tmp->size);

      if (strcmp(sec_tmp->name, "BANKHEADER") == 0)
	bankheader_status = ON;

      sec_tmp->i = 0;
      section_status = ON;

      continue;

    /* ENDS */

    case 's':
      section_status = OFF;
      bankheader_status = OFF;
      continue;

    /* DSB & DSW */

    case 'x':
      fscanf(file_out_ptr, "%d %d", &ind, &x);
      while (ind > 0) {
	if (mem_insert(x) == FAILED)
	  return FAILED;
	ind--;
      }
      continue;

    case 'X':
      fscanf(file_out_ptr, "%d %d", &ind, &inz);
      i = inz & 0xFF;
      inz = (inz >> 8) & 0xFF;

      while (ind > 0) {
	if (mem_insert(i) == FAILED)
	  return FAILED;
	if (mem_insert(inz) == FAILED)
	  return FAILED;
	ind--;
      }

      continue;

    /* DATA & OPTCODE */

    case 'd':
      fscanf(file_out_ptr, "%d", &x);
      if (mem_insert(x) == FAILED)
	return FAILED;
      continue;

    case 'y':
      fscanf(file_out_ptr, "%d", &inz);
      x = inz & 0xFF;
      inz = (inz >> 8) & 0xFF;

      if (mem_insert(x) == FAILED)
	return FAILED;
      if (mem_insert(inz) == FAILED)
	return FAILED;

      continue;

#ifdef W65816
    case 'b':
      fscanf(file_out_ptr, "%d", &base);
      continue;

    case 'z':
      fscanf(file_out_ptr, "%d", &inz);
      x = inz & 0xFF;
      ind = (inz >> 8) & 0xFF;
      inz = (inz >> 16) & 0xFF;

      if (mem_insert(x) == FAILED)
	return FAILED;
      if (mem_insert(ind) == FAILED)
	return FAILED;
      if (mem_insert(inz) == FAILED)
	return FAILED;

      continue;
#endif

    /* DATA BLOCK */

    case 'D':
      fscanf(file_out_ptr, "%d %d %d %d", &x, &inz, &z, &y);

      ifd_tmp = incbin_file_data_first;
      for (ind = 0; ind != x; ind++)
	ifd_tmp = ifd_tmp->next;
      t = ifd_tmp->data + z;

      /* swap? */

      if (inz == 1) {
	inz = y / 2;
	for (ind = 0; ind < inz; ind++) {
	  if (mem_insert(*(t + 1)) == FAILED)
	    return FAILED;
	  if (mem_insert(*t) == FAILED)
	    return FAILED;
	  t += 2;
	}
      }
      else {
	inz = y;
	for (ind = 0; ind < inz; ind++) {
	  if (mem_insert(*(t++)) == FAILED)
	    return FAILED;
	}
      }

      continue;

    /* ORIGIN & ROM BANK */

    case 'O':
    case 'B':

      if (c == 'O')
	fscanf(file_out_ptr, "%u", &pc_bank);
      else {
	fscanf(file_out_ptr, "%u %u", &rom_bank, &slot);
	if (banksize_defined == 0)
	  banksize = banks[rom_bank];
      }

      pc_full = bankaddress[rom_bank] + pc_bank;
      pc_slot = slots[slot].address + pc_bank;
      pc_slot_max = slots[slot].address + slots[slot].size;

      continue;

    /* LABEL */

    case 'L':
      fscanf(file_out_ptr, "%*s"); /* skip the label */
      continue;

    /* 8BIT COMPUTATION */

    case 'c':
      fscanf(file_out_ptr, "%d", &inz);

      if (bankheader_status == OFF)
	stacks_tmp = stacks_first;
      else
	stacks_tmp = stacks_header_first;

      for (ind = 0; ind != inz; ind++)
	stacks_tmp = stacks_tmp->next;

      stacks_tmp->section_status = section_status;
      if (section_status == ON) {
	stacks_tmp->section_id = sec_tmp->id;
	stacks_tmp->address = sec_tmp->i; /* relative address, to the beginning of the section */
      }
      else
	stacks_tmp->address = pc_bank; /* complete address, in ROM memory */

      stacks_tmp->bank = rom_bank;
      stacks_tmp->slot = slot;
      stacks_tmp->type = STACKS_TYPE_8BIT;

      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;

    /* 16BIT COMPUTATION */

    case 'C':
      fscanf(file_out_ptr, "%d", &inz);

      if (bankheader_status == OFF)
	stacks_tmp = stacks_first;
      else
	stacks_tmp = stacks_header_first;

      for (ind = 0; ind != inz; ind++)
	stacks_tmp = stacks_tmp->next;

      stacks_tmp->section_status = section_status;
      if (section_status == ON) {
	stacks_tmp->section_id = sec_tmp->id;
	stacks_tmp->address = sec_tmp->i; /* relative address, to the beginning of the section */
      }
      else
	stacks_tmp->address = pc_bank; /* complete address, in ROM memory */

      stacks_tmp->bank = rom_bank;
      stacks_tmp->slot = slot;
      stacks_tmp->type = STACKS_TYPE_16BIT;

      if (mem_insert_pad() == FAILED)
	return FAILED;
      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;

#ifdef W65816
    /* 24BIT COMPUTATION */

    case 'T':
      fscanf(file_out_ptr, "%d", &inz);

      if (bankheader_status == OFF)
	stacks_tmp = stacks_first;
      else
	stacks_tmp = stacks_header_first;

      for (ind = 0; ind != inz; ind++)
	stacks_tmp = stacks_tmp->next;

      stacks_tmp->section_status = section_status;
      if (section_status == ON) {
	stacks_tmp->section_id = sec_tmp->id;
	stacks_tmp->address = sec_tmp->i; /* relative address, to the beginning of the section */
      }
      else
	stacks_tmp->address = pc_bank; /* complete address, in ROM memory */

      stacks_tmp->bank = rom_bank;
      stacks_tmp->slot = slot;
      stacks_tmp->type = STACKS_TYPE_24BIT;
      stacks_tmp->base = base;

      if (mem_insert_pad() == FAILED)
	return FAILED;
      if (mem_insert_pad() == FAILED)
	return FAILED;
      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;

    /* 24BIT REFERENCE */

    case 'q':
      fscanf(file_out_ptr, "%256s", tmp);

      label_tmp = labels;
      x = 0;
      while (label_tmp != NULL) {
	if (strcmp(label_tmp->label, tmp) == 0) {

	  /* inside a section */
	  if (label_tmp->section_status == ON || tmp[0] == '_') {
	    if (new_unknown_reference(REFERENCE_TYPE_DIRECT_24BIT) == FAILED)
	      return FAILED;
	    if (mem_insert_pad() == FAILED)
	      return FAILED;
	    if (mem_insert_pad() == FAILED)
	      return FAILED;
	    if (mem_insert_pad() == FAILED)
	      return FAILED;
	    x = 1; /* section label */
	    label_tmp = NULL;
	    continue;
	  }

	  o = label_tmp->address + slots[label_tmp->slot].address;

	  if (mem_insert(o & 0xFF) == FAILED)
	    return FAILED;
	  if (mem_insert((o >> 8) & 0xFF) == FAILED)
	    return FAILED;
	  if (mem_insert(get_snes_cpu_bank(label_tmp)) == FAILED)
	    return FAILED;

	  x = 1;
	  label_tmp = NULL;
	  continue;
	}
	label_tmp = label_tmp->next;
      }

      /* try definitions */
      if (x == 0) {
	tmp_def = defines;
	while (tmp_def != NULL) {
	  if (strcmp(tmp, tmp_def->alias) == 0) {
	    if (tmp_def->type == DEFINITION_TYPE_STRING) {
	      if (error_level == ERROR_LEVEL_MAX)
		fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n", get_file_name(filename_id), line_number, tmp);
	      else
		fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", tmp);
	      return FAILED;
	    }
	    o = tmp_def->value;
	    x = 1;

	    if (mem_insert(o & 0xFF) == FAILED)
	      return FAILED;
	    if (mem_insert((o >> 8) & 0xFF) == FAILED)
	      return FAILED;
	    if (mem_insert((o >> 16) & 0xFF) == FAILED)
	      return FAILED;

	    break;
	  }
	  tmp_def = tmp_def->next;
	}
      }

      if (x == 1)
	continue;

      if (new_unknown_reference(REFERENCE_TYPE_DIRECT_24BIT) == FAILED)
	return FAILED;

      if (mem_insert_pad() == FAILED)
	return FAILED;
      if (mem_insert_pad() == FAILED)
	return FAILED;
      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;
#endif

    /* 16BIT REFERENCE */

    case 'r':
      fscanf(file_out_ptr, "%256s", tmp);

      label_tmp = labels;
      x = 0;
      while (label_tmp != NULL) {
	if (strcmp(label_tmp->label, tmp) == 0) {

	  /* inside a section */
	  if (label_tmp->section_status == ON || tmp[0] == '_') {
	    if (new_unknown_reference(REFERENCE_TYPE_DIRECT_16BIT) == FAILED)
	      return FAILED;
	    if (mem_insert_pad() == FAILED)
	      return FAILED;
	    if (mem_insert_pad() == FAILED)
	      return FAILED;
	    x = 1; /* section label */
	    label_tmp = NULL;
	    continue;
	  }

	  o = label_tmp->address + slots[label_tmp->slot].address;

	  if (mem_insert(o & 0xFF) == FAILED)
	    return FAILED;
	  if (mem_insert((o & 0xFF00) >> 8) == FAILED)
	    return FAILED;

	  x = 1;
	  label_tmp = NULL;
	  continue;
	}
	label_tmp = label_tmp->next;
      }

      /* try definitions */
      if (x == 0) {
	tmp_def = defines;
	while (tmp_def != NULL) {
	  if (strcmp(tmp, tmp_def->alias) == 0) {
	    if (tmp_def->type == DEFINITION_TYPE_STRING) {
	      if (error_level == ERROR_LEVEL_MAX)
		fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n", get_file_name(filename_id), line_number, tmp);
	      else
		fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", tmp);
	      return FAILED;
	    }
	    o = tmp_def->value;
	    x = 1;

	    if (mem_insert(o & 0xFF) == FAILED)
	      return FAILED;
	    if (mem_insert((o & 0xFF00) >> 8) == FAILED)
	      return FAILED;

	    break;
	  }
	  tmp_def = tmp_def->next;
	}
      }

      if (x == 1)
	continue;

      if (new_unknown_reference(REFERENCE_TYPE_DIRECT_16BIT) == FAILED)
	return FAILED;

      if (mem_insert_pad() == FAILED)
	return FAILED;
      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;

    /* 8BIT PC RELATIVE REFERENCE */

    case 'R':
      fscanf(file_out_ptr, "%256s", tmp);

      label_tmp = labels;
      x = 0;
      while (label_tmp != NULL) {
	if (strcmp(label_tmp->label, tmp) == 0) {

	  /* label is inside a section */
	  if (label_tmp->section_status == ON || tmp[0] == '_') {
	    if (new_unknown_reference(REFERENCE_TYPE_RELATIVE) == FAILED)
	      return FAILED;
	    if (mem_insert_pad() == FAILED)
	      return FAILED;
	    x = 1;
	    label_tmp = NULL;
	    continue;
	  }

	  /* label is outside a section */
	  o = label_tmp->address + slots[label_tmp->slot].address;
	  x = o - pc_slot - 1;

	  if (x > 128 || x < -127) {
	    if (error_level == ERROR_LEVEL_MAX)
	      fprintf(stderr, "%s:%d: INTERNAL_PASS_2: A PC relative reference to %s (%d) exceeds byte boundary.\n", get_file_name(filename_id), line_number, tmp, x);
	    else
	      fprintf(stderr, "INTERNAL_PASS_2: A PC relative reference to %s (%d) exceeds byte boundary.\n", tmp, x);
	    return FAILED;
	  }

	  if (mem_insert(x & 0xFF) == FAILED)
	    return FAILED;

	  x = 1;
	  label_tmp = NULL;
	  continue;
	}
	label_tmp = label_tmp->next;
      }

      /* try labels */
      if (x == 0) {
	tmp_def = defines;
	while (tmp_def != NULL) {
	  if (strcmp(tmp, tmp_def->alias) == 0) {
	    if (tmp_def->type == DEFINITION_TYPE_STRING) {
	      if (error_level == ERROR_LEVEL_MAX)
		fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n", get_file_name(filename_id), line_number, tmp);
	      else
		fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", tmp);
	      return FAILED;
	    }
	    o = tmp_def->value;
	    x = 1;

	    if (mem_insert(o & 0xFF) == FAILED)
	      return FAILED;

	    break;
	  }
	  tmp_def = tmp_def->next;
	}
      }

      if (x == 1)
	continue;

      if (new_unknown_reference(REFERENCE_TYPE_RELATIVE) == FAILED)
	return FAILED;

      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;

    /* 8BIT REFERENCE */

    case 'Q':
      fscanf(file_out_ptr, "%256s", tmp);

      /* try labels */
      x = 0;
      tmp_def = defines;
      while (tmp_def != NULL) {
	if (strcmp(tmp, tmp_def->alias) == 0) {
	  if (tmp_def->type == DEFINITION_TYPE_STRING) {
	    if (error_level == ERROR_LEVEL_MAX)
	      fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n", get_file_name(filename_id), line_number, tmp);
	    else
	      fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", tmp);
	    return FAILED;
	  }
	  o = tmp_def->value;
	  x = 1;

	  if (mem_insert(o & 0xFF) == FAILED)
	    return FAILED;

	  break;
	}
	tmp_def = tmp_def->next;
      }

      if (x == 1)
	continue;

      if (new_unknown_reference(REFERENCE_TYPE_DIRECT_8BIT) == FAILED)
	return FAILED;

      if (mem_insert_pad() == FAILED)
	return FAILED;

      continue;
    }
  }

  fclose(file_out_ptr);
  file_out_ptr = NULL;

  /* do compact mode changes */

  if (compact_mode == ON) {

    lb[0] = '_';
    lb[2] = 0;
    symb_size = strlen(symb);

    label_tmp = labels;
    while (label_tmp != NULL) {
      if (label_tmp->label[0] == '_') {

	/* get a new name */
	while (compact_mode_next_name() == FAILED);

	/* replace references */
	tul = unknown_labels;
	while (tul != NULL) {
	  if (strcmp(tul->label, label_tmp->label) == 0)
	    strcpy(tul->label, lb);
	  tul = tul->next;
	}

	/* replace label */
	strcpy(label_tmp->label, lb);
      }
      label_tmp = label_tmp->next;
    }
  }

  /* rom/program file output */
  if (output_format == OUTPUT_ROM || output_format == OUTPUT_PROGRAM) {

    if (export_first != NULL)
      fprintf(stderr, "WARNING: .EXPORT's are not used when compiling a ROM/PROGRAM file.\n");

    /* output sections */
    if (output_section(SECTION_STATUS_FORCE) == FAILED)
      return FAILED;
    if (output_section(SECTION_STATUS_ABSOLUTE) == FAILED)
      return FAILED;
    if (output_section(SECTION_STATUS_FREE) == FAILED) /* take care of semifree sections as well */
      return FAILED;
    if (output_section(SECTION_STATUS_OVERWRITE) == FAILED)
      return FAILED;

    /* fix references to labels */
    mem_insert_overwrite = OFF;

    tul = unknown_labels;
    while (tul != NULL) {

      /* find source address */
      if (tul->section_status == ON) {

	/* get section address */
	sec_tmp = sections_first;
	while (tul->section_id != sec_tmp->id)
	  sec_tmp = sec_tmp->next;

	if (sec_tmp->status != SECTION_STATUS_ABSOLUTE) {
	  pc_bank = tul->address + sec_tmp->address;
	  pc_full = pc_bank + bankaddress[sec_tmp->bank];
	  if (banksize_defined == 0)
	    banksize = banks[sec_tmp->bank];

	  /* get slot address */
	  pc_slot = slots[sec_tmp->slot].address + pc_bank;
	  pc_slot_max = slots[sec_tmp->slot].address + slots[sec_tmp->slot].size;
	}
	else {
	  pc_bank = 0;
	  if (banksize_defined == 0)
	    banksize = 0x10000;
	  pc_slot = 0;
	  pc_slot_max = 0x10000;
	  pc_full = sec_tmp->address + tul->address;
	}
      }
      else {
	pc_full = tul->address + bankaddress[tul->bank];
	pc_bank = tul->address;
	if (banksize_defined == 0)
	  banksize = banks[tul->bank];

	/* get slot address */
	pc_slot = slots[tul->slot].address + pc_bank;
	pc_slot_max = slots[tul->slot].address + slots[tul->slot].size;
      }

      /* look through labels */
      label_tmp = labels;
      while (label_tmp != NULL) {
	if (strcmp(label_tmp->label, tul->label) == 0) {
	  if (tul->label[0] == '_' && tul->section_status != label_tmp->section_status) {
	    label_tmp = label_tmp->next;
	    continue;
	  }
	  if (tul->label[0] == '_' && tul->section_status == ON && tul->section_id != label_tmp->section_id) {
	    label_tmp = label_tmp->next;
	    continue;
	  }

	  if (label_tmp->section_status == ON) {
	    sec_tmp = sections_first;
	    while (sec_tmp != NULL) {
	      if (label_tmp->section_id == sec_tmp->id)
		break;
	      else
		sec_tmp = sec_tmp->next;
	    }
	    inz = sec_tmp->address + label_tmp->address + slots[sec_tmp->slot].address;
	  }
	  else
	    inz = label_tmp->address + slots[label_tmp->slot].address;

	  break;
	}
	else
	  label_tmp = label_tmp->next;
      }

#ifdef W65816
      /* a 24bit address label? */
      if (label_tmp != NULL && tul->type == REFERENCE_TYPE_DIRECT_24BIT)
	inz += (get_snes_cpu_bank(label_tmp)) << 16;
      /* not a label? search definitions */
      else {
#else
      if (label_tmp == NULL) {
#endif
	tmp_def = defines;
	while (tmp_def != NULL) {
	  if (strcmp(tul->label, tmp_def->alias) == 0) {
	    if (tmp_def->type == DEFINITION_TYPE_STRING) {
	      if (error_level == ERROR_LEVEL_MAX)
		fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n", get_file_name(tul->filename_id), tul->linenumber, tul->label);
	      else
		fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", tul->label);
	      return FAILED;
	    }
	    inz = tmp_def->value;
	    label_tmp = (struct label_def *)1;
	    break;
	  }
	  tmp_def = tmp_def->next;
	}
      }

      if (label_tmp == NULL) {
	if (error_level == ERROR_LEVEL_MAX)
	  fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", get_file_name(tul->filename_id), tul->linenumber, tul->label);
	else
	  fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", tul->label);
	return FAILED;
      }

      /* addresses calculated, insert data */
      if (tul->type == REFERENCE_TYPE_DIRECT_16BIT) {
	if (mem_insert_fix(inz & 0xFF) == FAILED)
	  return FAILED;
	if (mem_insert_fix((inz >> 8) & 0xFF) == FAILED)
	  return FAILED;
      }
      else if (tul->type == REFERENCE_TYPE_DIRECT_8BIT) {
	if (mem_insert_fix(inz & 0xFF) == FAILED)
	  return FAILED;
      }
#ifdef W65816
      else if (tul->type == REFERENCE_TYPE_DIRECT_24BIT) {
	if (mem_insert_fix(inz & 0xFF) == FAILED)
	  return FAILED;
	if (mem_insert_fix((inz >> 8) & 0xFF) == FAILED)
	  return FAILED;
	if (mem_insert_fix((inz >> 16) & 0xFF) == FAILED)
	  return FAILED;
      }
#endif
      else {
	x = inz - pc_slot - 1;
	if (x > 255 || x < -127) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: A PC relative reference to %s (%d) exceeds byte boundary.\n", get_file_name(tul->filename_id), tul->linenumber, tul->label, x);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: A PC relative reference to %s (%d) exceeds byte boundary.\n", tul->label, x);
	  return FAILED;
	}
	if (mem_insert_fix(x & 0xFF) == FAILED)
	  return FAILED;
      }

      tul = tul->next;
    }

    /* fix references to labels inside bank header sections */
    mem_insert_overwrite = OFF;

    tul = unknown_header_labels;
    while (tul != NULL) {

      unsigned char *dat;


      sec_tmp = sections_first;
      while (tul->section_id != sec_tmp->id)
	sec_tmp = sec_tmp->next;
      dat = sec_tmp->data + tul->address;

      label_tmp = labels;
      while (label_tmp != NULL) {
	if (strcmp(label_tmp->label, tul->label) == 0) {
	  if (label_tmp->section_status == ON) {
	    if (tul->section_status == ON) {
	      /* is it in the same section, if local? */
	      if (label_tmp->section_id != tul->section_id && tul->label[0] == '_') {
		label_tmp = label_tmp->next;
		continue;
	      }
	    }
	    sec_tmp = sections_first;
	    while (sec_tmp != NULL) {
	      if (label_tmp->section_status == ON && label_tmp->section_id == sec_tmp->id)
		break;
	      else
		sec_tmp = sec_tmp->next;
	    }
	    inz = sec_tmp->address + label_tmp->address + slots[sec_tmp->slot].address;
	  }
	  else
	    inz = label_tmp->address + slots[label_tmp->slot].address;

	  break;
	}
	else
	  label_tmp = label_tmp->next;
      }

#ifdef W65816
      /* a 24bit address label? */
      if (label_tmp != NULL && tul->type == REFERENCE_TYPE_DIRECT_24BIT)
	inz += (get_snes_cpu_bank(label_tmp)) << 16;
      /* not a label? search definitions */
      else {
#else
      if (label_tmp == NULL) {
#endif
	tmp_def = defines;
	while (tmp_def != NULL) {
	  if (strcmp(tul->label, tmp_def->alias) == 0) {
	    if (tmp_def->type == DEFINITION_TYPE_STRING) {
	      if (error_level == ERROR_LEVEL_MAX)
		fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n",
			get_file_name(tul->filename_id), tul->linenumber, tul->label);
	      else
		fprintf(stderr, "INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n", tul->label);
	      return FAILED;
	    }
	    inz = tmp_def->value;
	    label_tmp = (struct label_def *)1;
	    break;
	  }
	  tmp_def = tmp_def->next;
	}
      }

      if (label_tmp == NULL) {
	if (error_level == ERROR_LEVEL_MAX)
	  fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", get_file_name(tul->filename_id), tul->linenumber, tul->label);
	else
	  fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", tul->label);
	return FAILED;
      }

      /* addresses calculated, insert data */
      if (tul->type == REFERENCE_TYPE_DIRECT_16BIT) {
	*dat = inz & 0xFF;
	dat++;
	*dat = (inz >> 8) & 0xFF;
      }
#ifdef W65816
      else if (tul->type == REFERENCE_TYPE_DIRECT_24BIT) {
	*dat = inz & 0xFF;
	dat++;
	*dat = (inz >> 8) & 0xFF;
	dat++;
	*dat = (inz >> 16) & 0xFF;
      }
#endif
      else {
	if (inz > 255 || inz < -127) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: An 8bit reference to %s (%d) exceeds byte boundary.\n",
		    get_file_name(tul->filename_id), tul->linenumber, tul->label, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: An 8bit reference to %s (%d) exceeds byte boundary.\n", tul->label, inz);
	  return FAILED;
	}
	*dat = inz;
      }

      tul = tul->next;
    }

    /* finish all pending stack calculations */
    stacks_tmp = stacks_first;
    while (stacks_tmp != NULL) {

      struct stackitem *sta;
      int g;


      /* find source address */
      if (stacks_tmp->section_status == ON) {

	/* get section address */
	sec_tmp = sections_first;
	while (stacks_tmp->section_id != sec_tmp->id)
	  sec_tmp = sec_tmp->next;

	if (sec_tmp->status != SECTION_STATUS_ABSOLUTE) {
	  pc_bank = stacks_tmp->address + sec_tmp->address;
	  pc_full = pc_bank + bankaddress[sec_tmp->bank];
	  if (banksize_defined == 0)
	    banksize = banks[sec_tmp->bank];

	  /* get slot address */
	  pc_slot = slots[sec_tmp->slot].address + pc_bank;
	  pc_slot_max = slots[sec_tmp->slot].address + slots[sec_tmp->slot].size;
	}
	else {
	  pc_bank = 0;
	  if (banksize_defined == 0)
	    banksize = 0x10000;
	  pc_slot = 0;
	  pc_slot_max = 0x10000;
	  pc_full = sec_tmp->address + tul->address;
	}
      }
      else {
	pc_bank = stacks_tmp->address;
	pc_full = pc_bank + bankaddress[stacks_tmp->bank];
	if (banksize_defined == 0)
	  banksize = banks[stacks_tmp->bank];

	/* get slot address */
	pc_slot = slots[stacks_tmp->slot].address + pc_bank;
	pc_slot_max = slots[stacks_tmp->slot].address + slots[stacks_tmp->slot].size;
      }

      sta = stacks_tmp->stack;
      g = 0;
      while (g != stacks_tmp->stacksize) {
	if (sta->type == STACK_ITEM_TYPE_STRING) {

	  label_tmp = labels;
	  while (label_tmp != NULL) {
	    if (strcmp(label_tmp->label, sta->string) == 0) {
	      if (label_tmp->section_status == ON) {
		if (stacks_tmp->section_status == ON) {
		  /* is it in the same section, if local? */
		  if (label_tmp->section_id != stacks_tmp->section_id && sta->string[0] == '_') {
		    label_tmp = label_tmp->next;
		    continue;
		  }
		}
		sec_tmp = sections_first;
		while (sec_tmp != NULL) {
		  if (label_tmp->section_status == ON && label_tmp->section_id == sec_tmp->id)
		    break;
		  else
		    sec_tmp = sec_tmp->next;
		}
		inz = sec_tmp->address + label_tmp->address + slots[sec_tmp->slot].address;
	      }
	      else
		inz = label_tmp->address + slots[label_tmp->slot].address;

	      break;
	    }
	    else
	      label_tmp = label_tmp->next;
	  }

#ifdef W65816
	  /* a 24bit address label? */
	  if (label_tmp != NULL && stacks_tmp->type == STACKS_TYPE_24BIT)
	    inz += (get_snes_cpu_bank(label_tmp)) << 16;
	  /* not a label? search definitions */
	  else {
#else
	  if (label_tmp == NULL) {
#endif
	    tmp_def = defines;
	    while (tmp_def != NULL) {
	      if (strcmp(sta->string, tmp_def->alias) == 0) {
		if (tmp_def->type == DEFINITION_TYPE_STRING) {
		  if (error_level == ERROR_LEVEL_MAX)
		    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n",
			    get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, sta->string);
		  else
		    fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", sta->string);
		  return FAILED;
		}
		inz = tmp_def->value;
		label_tmp = (struct label_def *)1;
		break;
	      }
	      tmp_def = tmp_def->next;
	    }
	  }

	  if (label_tmp == NULL) {
	    if (error_level == ERROR_LEVEL_MAX)
	      fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Unresolved reference to \"%s\".\n",
		      get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, sta->string);
	    else
	      fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", sta->string);
	    return FAILED;
	  }

	  sta->value = inz;
	  sta->type = STACK_ITEM_TYPE_VALUE;
	}
	sta++;
	g++;
      }

      /* all the references have been decoded, now compute */
      if (compute_stack(stacks_tmp, stacks_tmp->stacksize, &inz) == FAILED)
	return FAILED;

      if (stacks_tmp->type == STACKS_TYPE_8BIT) {
	if (inz > 255 || inz < -127) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Expected a 8bit value, got %d.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Expected a 8bit value, got %d.\n", inz);
	  return FAILED;
	}
	if (mem_insert_fix(inz) == FAILED)
	  return FAILED;
      }
      else if (stacks_tmp->type == STACKS_TYPE_16BIT) {
	if (inz > 65535 || inz < -32768) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Expected a 16bit value, got %d.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Expected a 16bit value, got %d.\n", inz);
	  return FAILED;
	}
	if (mem_insert_fix(inz & 0xFF) == FAILED)
	  return FAILED;
	if (mem_insert_fix((inz >> 8) & 0xFF) == FAILED)
	  return FAILED;
      }
#ifdef W65816
      else {
	if (inz > 16777215 || inz < -8388608) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Expected a 24bit value, got %d.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Expected a 24bit value, got %d.\n", inz);
	  return FAILED;
	}
	if (mem_insert_fix(inz & 0xFF) == FAILED)
	  return FAILED;
	if (mem_insert_fix((inz >> 8) & 0xFF) == FAILED)
	  return FAILED;
	if (mem_insert_fix((inz >> 16) & 0xFF) == FAILED)
	  return FAILED;
      }
#endif

      /* next stack computation */
      stacks_tmp = stacks_tmp->next;
    }

    /* finish all pending stack calculations inside bank headers */
    stacks_tmp = stacks_header_first;
    while (stacks_tmp != NULL) {

      struct stackitem *sta;
      unsigned char *dat;
      int g;


      /* get section address */
      sec_tmp = sections_first;
      while (stacks_tmp->section_id != sec_tmp->id)
	sec_tmp = sec_tmp->next;
      dat = sec_tmp->data + stacks_tmp->address;

      sta = stacks_tmp->stack;
      g = 0;
      while (g != stacks_tmp->stacksize) {
	if (sta->type == STACK_ITEM_TYPE_STRING) {

	  label_tmp = labels;
	  while (label_tmp != NULL) {
	    if (strcmp(label_tmp->label, sta->string) == 0) {
	      if (label_tmp->section_status == ON) {
		if (stacks_tmp->section_status == ON) {
		  /* is it in the same section, if local? */
		  if (label_tmp->section_id != stacks_tmp->section_id && sta->string[0] == '_') {
		    label_tmp = label_tmp->next;
		    continue;
		  }
		}
		sec_tmp = sections_first;
		while (sec_tmp != NULL) {
		  if (label_tmp->section_status == ON && label_tmp->section_id == sec_tmp->id)
		    break;
		  else
		    sec_tmp = sec_tmp->next;
		}
		if (sec_tmp == NULL) {
		  if (error_level == ERROR_LEVEL_MAX)
		    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: You can't refer to labels inside BANKHEADER sections.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber);
		  else
		    fprintf(stderr, "INTERNAL_PASS_2: You can't refer to labels inside BANKHEADER sections.\n");
		  return FAILED;
		}
		inz = sec_tmp->address + label_tmp->address + slots[sec_tmp->slot].address;
	      }
	      else
		inz = label_tmp->address + slots[label_tmp->slot].address;

	      break;
	    }
	    else
	      label_tmp = label_tmp->next;
	  }

#ifdef W65816
	  /* a 24bit address label? */
	  if (label_tmp != NULL && stacks_tmp->type == STACKS_TYPE_24BIT)
	    inz += (get_snes_cpu_bank(label_tmp)) << 16;
	  /* not a label? search definitions */
	  else {
#else
	  if (label_tmp == NULL) {
#endif
	    tmp_def = defines;
	    while (tmp_def != NULL) {
	      if (strcmp(sta->string, tmp_def->alias) == 0) {
		if (tmp_def->type == DEFINITION_TYPE_STRING) {
		  if (error_level == ERROR_LEVEL_MAX)
		    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Reference to a string definition \"%s\"?\n",
			    get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, sta->string);
		  else
		    fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to a string definition \"%s\"?\n", sta->string);
		  return FAILED;
		}
		inz = tmp_def->value;
		label_tmp = (struct label_def *)1;
		break;
	      }
	      tmp_def = tmp_def->next;
	    }
	  }

	  if (label_tmp == NULL) {
	    if (error_level == ERROR_LEVEL_MAX)
	      fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, sta->string);
	    else
	      fprintf(stderr, "INTERNAL_PASS_2: Unresolved reference to \"%s\".\n", sta->string);
	    return FAILED;
	  }

	  sta->value = inz;
	  sta->type = STACK_ITEM_TYPE_VALUE;
	}
	sta++;
	g++;
      }

      /* all the references have been decoded, now compute */
      if (compute_stack(stacks_tmp, stacks_tmp->stacksize, &inz) == FAILED)
	return FAILED;

      if (stacks_tmp->type == STACKS_TYPE_8BIT) {
	if (inz > 255 || inz < -127) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Expected a 8bit value, got %d.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Expected a 8bit value, got %d.\n", inz);
	  return FAILED;
	}
	*dat = inz & 0xFF;
      }
      else if (stacks_tmp->type == STACKS_TYPE_16BIT) {
	if (inz > 65535 || inz < -32768) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Expected a 16bit value, got %d.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Expected a 16bit value, got %d.\n", inz);
	  return FAILED;
	}
	*dat = inz & 0xFF;
	dat++;
	*dat = (inz >> 8) & 0xFF;
      }
#ifdef W65816
      else {
	if (inz > 16777215 || inz < -8388608) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: INTERNAL_PASS_2: Expected a 24bit value, got %d.\n", get_file_name(stacks_tmp->filename_id), stacks_tmp->linenumber, inz);
	  else
	    fprintf(stderr, "INTERNAL_PASS_2: Expected a 24bit value, got %d.\n", inz);
	  return FAILED;
	}
	*dat = inz & 0xFF;
	dat++;
	*dat = (inz >> 8) & 0xFF;
	dat++;
	*dat = (inz >> 16) & 0xFF;
      }
#endif

      /* next stack computation */
      stacks_tmp = stacks_tmp->next;
    }

    /* overwriting not possible */
    mem_insert_overwrite = OFF;

#ifdef GB
    /* compute complement check */
    if (computecomplementcheck_defined != 0) {
      ind = 0;
      for (x = 0x134; x <= 0x14C; x++)
	ind += rom_banks[x];
      ind += 25;

      mem_insert_absolute(0x14D, (0 - (ind & 0xFF)) & 0xFF);
    }

    /* compute cheksum */
    if (computechecksum_defined != 0) {
      ind = 0;
      for (x = 0; x < 0x14E; x++)
	ind += rom_banks[x];
      for (x = 0x150; x < max_address; x++)
	ind += rom_banks[x];

      mem_insert_absolute(0x14E, (ind & 0xFF00) >> 8);
      mem_insert_absolute(0x14F, ind & 0xFF);
    }
#endif

#ifdef Z80
    if (computesmschecksum_defined != 0) {
      ind = 0;
      for (x = 0; x < 0x7FF0; x++)
	ind += rom_banks[x];

      if (max_address > (256 * 1024))
	inz = 256 * 1024;
      else
	inz = max_address;

      for (x = 0x8000; x < inz; x++)
	ind += rom_banks[x];

      mem_insert_absolute(0x7FFA, ind & 0xFF);
      mem_insert_absolute(0x7FFB, (ind & 0xFF00) >> 8);
    }
#endif

    /* write data */
    if (test_mode == OFF) {

      if ((final_ptr = fopen(final_name, "wb")) == NULL) {
	fprintf(stderr, "INTERNAL_PASS_2: Error opening file \"%s\".\n", final_name);
	return FAILED;
      }

#ifndef GB
      if (header_defined != 0) {
	ifd_tmp = incbin_file_data_first;
	for (ind = 0; ind != header; ind++)
	  ifd_tmp = ifd_tmp->next;
	fwrite(ifd_tmp->data, 1, ifd_tmp->size, final_ptr);
	header_size = ifd_tmp->size;
      }

      /* output a program */
      if (output_format == OUTPUT_PROGRAM) {
	for (ind = 0; rom_banks_usage_table[ind] == 0 && ind < max_address; ind++);
	inz = ind;
	binary_start = ind;
	while (inz < max_address) {
	  if (rom_banks_usage_table[inz] > 0)
	    x = inz;
	  inz++;
	}
	final_size = x - ind + 1;

	/* find bank 0 header section */
	sec_tmp = sections_first;
	while (sec_tmp != NULL) {
	  if (sec_tmp->status == SECTION_STATUS_HEADER && sec_tmp->bank == 0) {
	    fwrite(sec_tmp->data, sec_tmp->size, 1, final_ptr);
	    break;
	  }
	  sec_tmp = sec_tmp->next;
	}

	fwrite(&rom_banks[ind], final_size, 1, final_ptr);
      }

      if (footer_defined != 0) {
	ifd_tmp = incbin_file_data_first;
	for (ind = 0; ind != footer; ind++)
	  ifd_tmp = ifd_tmp->next;
	fwrite(ifd_tmp->data, 1, ifd_tmp->size, final_ptr);
	footer_size = ifd_tmp->size;
      }
#endif

      /* output rom banks with possible headers */
      if (output_format == OUTPUT_ROM) {
	for (ind = 0; ind < rombanks; ind++) {
	  sec_tmp = sections_first;
	  while (sec_tmp != NULL) {
	    if (sec_tmp->status == SECTION_STATUS_HEADER && sec_tmp->bank == ind) {
	      fwrite(sec_tmp->data, sec_tmp->size, 1, final_ptr);
	      break;
	    }
	    sec_tmp = sec_tmp->next;
	  }
	  fwrite(rom_banks + bankaddress[ind], banks[ind], 1, final_ptr);
	}
      }

      fclose(final_ptr);
    }

#ifdef GB
    /* check RAM banks */
    if (rom_banks_usage_table[329] == 0)
      fprintf(stderr, "WARNING: The amount of RAM banks wasn't defined.\n");
#endif

    /* export symbolic information file */
    if (symbol_mode == ON) {

      char *p;


      p = final_name;
      for (ind = 0; ind < MAX_NAME_LENGTH && *p != '.' && *p != 0x0; ind++, p++);
      *p++ = '.';
      *p++ = 's';
      *p++ = 'y';
      *p++ = 'm';
      *p = 0x0;

      if ((final_ptr = fopen(final_name, "wb")) == NULL) {
	fprintf(stderr, "INTERNAL_PASS_2: Error opening file \"%s\".\n", final_name);
	return FAILED;
      }

      fprintf(final_ptr, "; this file was created with wla - yet another\n");
      fprintf(final_ptr, "; gb-z80/z80/6502/6510/65816 macro assembler by\n");
      fprintf(final_ptr, "; ville helin <vhelin@cc.hut.fi>.\n");

      label_tmp = labels;
      while (label_tmp != NULL) {
	ind = label_tmp->address;
	if (label_tmp->section_status == ON) {
	  sec_tmp = sections_first;
	  while (label_tmp->section_id != sec_tmp->id)
	    sec_tmp = sec_tmp->next;
	  ind += sec_tmp->address;
	  inz = sec_tmp->slot;
	}
	ind += slots[label_tmp->slot].address;
	fprintf(final_ptr, "%.4x:%.4x %s\n", label_tmp->bank, ind, label_tmp->label);
	label_tmp = label_tmp->next;
      }

      fclose(final_ptr);
    }
  }

  /* library file output */
  else if (output_format == OUTPUT_LIBRARY && test_mode == OFF) {

    if ((final_ptr = fopen(final_name, "wb")) == NULL) {
      fprintf(stderr, "INTERNAL_PASS_2: Error opening file \"%s\".\n", final_name);
      return FAILED;
    }

    /* header */
    fprintf(final_ptr, "WLAP");                   /* format information       */

    /* export definitions */
    if (export_definitions(final_ptr) == FAILED)
      return FAILED;

    /* labels */
    ov = 0;
    label_tmp = labels;
    while (label_tmp != NULL) {
      ov++;
      label_tmp = label_tmp->next;
    }

    WRITEOUT_OV

    label_tmp = labels;
    while (label_tmp != NULL) {
      fprintf(final_ptr, "%s%c", label_tmp->label, 0x0);
      if (label_tmp->section_status == OFF) {
	fprintf(stderr, "INTERNAL_PASS_2: All labels must be inside sections in library output mode.\n");
	return FAILED;
      }
      else
	fprintf(final_ptr, "%c", label_tmp->section_id);

      ov = label_tmp->address;
      WRITEOUT_OV

      label_tmp = label_tmp->next;
    }

    /* unknown labels */
    ov = 0;
    label_tmp = unknown_labels;
    while (label_tmp != NULL) {
      ov++;
      label_tmp = label_tmp->next;
    }

    WRITEOUT_OV

    label_tmp = unknown_labels;
    while (label_tmp != NULL) {
      fprintf(final_ptr, "%s%c%c", label_tmp->label, 0x0, label_tmp->type);
      if (label_tmp->section_status == OFF) {
	fprintf(stderr, "INTERNAL_PASS_2: Label \"%s\" is outside all sections.\n", label_tmp->label);
	return FAILED;
      }
      else
	fprintf(final_ptr, "%c", label_tmp->section_id);

#ifndef W65816
      ov = label_tmp->address;
#else
      ov = label_tmp->address + (label_tmp->base << 16);
#endif
      WRITEOUT_OV

      label_tmp = label_tmp->next;
    }

    /* pending calculations */
    ov = stacks_outside;
    WRITEOUT_OV

    stacks_tmp = stacks_first;
    while (stacks_tmp != NULL) {
      fprintf(final_ptr, "%c%c", stacks_tmp->type, stacks_tmp->section_id);

      ov = stacks_tmp->address;
      WRITEOUT_OV

      ov = stacks_tmp->linenumber;
      WRITEOUT_OV

      fprintf(final_ptr, "%c", stacks_tmp->stacksize);

      for (ind = 0; ind < stacks_tmp->stacksize; ind++) {
	fprintf(final_ptr, "%c%c", stacks_tmp->stack[ind].type, stacks_tmp->stack[ind].sign);
	if (stacks_tmp->stack[ind].type == STACK_ITEM_TYPE_STRING)
	  fprintf(final_ptr, "%s%c", stacks_tmp->stack[ind].string, 0);
	else {
	  ov = stacks_tmp->stack[ind].value;
	  WRITEOUT_OV
	}
      }

      stacks_tmp = stacks_tmp->next;
    }

    /* sections */
    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      fprintf(final_ptr, "%s%c", sec_tmp->name, sec_tmp->status);

      ov = sec_tmp->size;
      WRITEOUT_OV

      fwrite(sec_tmp->data, 1, sec_tmp->size, final_ptr);

      sec_tmp = sec_tmp->next;
    }

    fclose(final_ptr);
  }

  /* object file output */
  else if (output_format == OUTPUT_OBJECT && test_mode == OFF) {

    if ((final_ptr = fopen(final_name, "wb")) == NULL) {
      fprintf(stderr, "INTERNAL_PASS_2: Error opening file \"%s\".\n", final_name);
      return FAILED;
    }

    /* header */

#ifdef GB
    fprintf(final_ptr, "WLA9");                   /* format information       */
    if (name_defined != 0)
      fprintf(final_ptr, "%c", 1);                /* rom name status          */
    else
      fprintf(final_ptr, "%c", 0);
    if (romgbc != 0)
      inz = 15;
    else
      inz = 16;
    for (ind = 0; ind < inz; ind++)
      fprintf(final_ptr, "%c", name[ind]);        /* rom name                 */
    if (romgbc != 0)
      fprintf(final_ptr, "%c", 0x80);             /* gbc indicator            */
    fprintf(final_ptr, "%c", romtype);            /* rom type                 */
    fprintf(final_ptr, "%c", rambanks);           /* ram size                 */
    fprintf(final_ptr, "%c", emptyfill);          /* empty fill               */
    fprintf(final_ptr, "%c", cartridgetype);      /* cartridge type           */
    if (licenseecodeold_defined != 0)
      fprintf(final_ptr, "%c", 0x0);              /* licensee code old in use */
    else
      fprintf(final_ptr, "%c", 0x1);              /* licensee code new in use */
    fprintf(final_ptr, "%c", licenseecodeold);    /* licensee code old        */
    fprintf(final_ptr, "%c", licenseecodenew_c1); /* licensee code new c1     */
    fprintf(final_ptr, "%c", licenseecodenew_c2); /* licensee code new c2     */
    if (computechecksum_defined != 0)
      fprintf(final_ptr, "%c", 0x1);              /* compute checksum         */
    else
      fprintf(final_ptr, "%c", 0x0);
    if (computecomplementcheck_defined != 0)
      fprintf(final_ptr, "%c", 0x1);              /* compute complement check */
    else
      fprintf(final_ptr, "%c", 0x0);
#endif

#ifndef GB
    fprintf(final_ptr, "WLAB");                   /* format information       */
    ov = rombanks;
    WRITEOUT_OV                                   /* amount of rom banks      */
    fprintf(final_ptr, "%c", emptyfill);          /* empty fill               */
#ifdef Z80
    if (computesmschecksum_defined != 0)
      fprintf(final_ptr, "%c", 1);                /* sms checksum */
    else
#endif
#ifdef W65816
    if (snes_mode != 0) {
      ind = 4; /* snes */
      if (hirom_defined != 0)
	ind += 2;
      if (fastrom_defined != 0)
	ind += 8;
      fprintf(final_ptr, "%c", ind);              /* snes tags */
    }
    else
#endif
      fprintf(final_ptr, "%c", 0);                /* no sms checksum */
#endif

    /* rom bank map */
    if (rombankmap_defined != 0) {
      fprintf(final_ptr, "%c", 1);                /* status */
      for (i = 0; i < rombanks; i++) {
	ov = banks[i];
	WRITEOUT_OV                               /* banksize */
      }
    }
    else {
      fprintf(final_ptr, "%c", 0);                /* status */
      ov = banksize;
      WRITEOUT_OV                                 /* banksize */
    }

    /* memory map */
    fprintf(final_ptr, "%c", slots_amount);       /* amount of slots */

    for (i = 0; i < slots_amount; i++) {
      if (slots[i].size != 0) {
	ov = slots[i].address;
	WRITEOUT_OV                               /* slot address */
	ov = slots[i].size;
	WRITEOUT_OV                               /* slot size */
      }
    }

    /* export definitions */
    if (export_definitions(final_ptr) == FAILED)
      return FAILED;

    /* labels */
    ov = 0;
    label_tmp = labels;
    while (label_tmp != NULL) {
      ov++;
      label_tmp = label_tmp->next;
    }

    /* amount of labels */
    WRITEOUT_OV

    label_tmp = labels;
    while (label_tmp != NULL) {
      fprintf(final_ptr, "%s%c", label_tmp->label, 0x0);
      if (label_tmp->section_status == OFF)
	fprintf(final_ptr, "%c", 0x0);
      else
	fprintf(final_ptr, "%c", label_tmp->section_id);

#ifndef W65816
      ov = label_tmp->address;
#else
      ov = label_tmp->address + (label_tmp->base << 16);
#endif
      WRITEOUT_OV

      ov = label_tmp->bank;
      WRITEOUT_OV

      fprintf(final_ptr, "%c", label_tmp->slot);

      label_tmp = label_tmp->next;
    }

    /* outside references */
    ov = 0;
    label_tmp = unknown_labels;
    while (label_tmp != NULL) {
      ov++;
      label_tmp = label_tmp->next;
    }
    label_tmp = unknown_header_labels;
    while (label_tmp != NULL) {
      ov++;
      label_tmp = label_tmp->next;
    }

    WRITEOUT_OV

    label_tmp = unknown_labels;
    while (label_tmp != NULL) {
      fprintf(final_ptr, "%s%c%c", label_tmp->label, 0x0, label_tmp->type);
      if (label_tmp->section_status == OFF)
	fprintf(final_ptr, "%c", 0x0);
      else
	fprintf(final_ptr, "%c", label_tmp->section_id);

      ov = label_tmp->address;
      WRITEOUT_OV

      ov = label_tmp->bank;
      WRITEOUT_OV

      fprintf(final_ptr, "%c", label_tmp->slot);

      label_tmp = label_tmp->next;
    }
    label_tmp = unknown_header_labels;
    while (label_tmp != NULL) {
      fprintf(final_ptr, "%s%c%c", label_tmp->label, 0x0, label_tmp->type);
      if (label_tmp->section_status == OFF)
	fprintf(final_ptr, "%c", 0x0);
      else
	fprintf(final_ptr, "%c", label_tmp->section_id);

      ov = label_tmp->address;
      WRITEOUT_OV

      ov = label_tmp->bank;
      WRITEOUT_OV

      fprintf(final_ptr, "%c", label_tmp->slot);

      label_tmp = label_tmp->next;
    }

    /* pending calculations */
    ov = stacks_outside + stacks_inside;
    WRITEOUT_OV

    stacks_tmp = stacks_first;
    while (stacks_tmp != NULL) {
      if (stacks_tmp->section_status == ON)
	fprintf(final_ptr, "%c%c", stacks_tmp->type, stacks_tmp->section_id);
      else
	fprintf(final_ptr, "%c%c", stacks_tmp->type, 0);

      ov = stacks_tmp->address;
      WRITEOUT_OV

      ov = stacks_tmp->linenumber;
      WRITEOUT_OV

      ov = stacks_tmp->bank;
      WRITEOUT_OV

      fprintf(final_ptr, "%c%c", stacks_tmp->slot, stacks_tmp->stacksize);

      for (ind = 0; ind < stacks_tmp->stacksize; ind++) {
	fprintf(final_ptr, "%c%c", stacks_tmp->stack[ind].type, stacks_tmp->stack[ind].sign);
	if (stacks_tmp->stack[ind].type == STACK_ITEM_TYPE_STRING)
	  fprintf(final_ptr, "%s%c", stacks_tmp->stack[ind].string, 0);
	else {
	  ov = stacks_tmp->stack[ind].value;
	  WRITEOUT_OV
	}
      }

      stacks_tmp = stacks_tmp->next;
    }
    stacks_tmp = stacks_header_first;
    while (stacks_tmp != NULL) {
      if (stacks_tmp->section_status == ON)
	fprintf(final_ptr, "%c%c", stacks_tmp->type, stacks_tmp->section_id);
      else
	fprintf(final_ptr, "%c%c", stacks_tmp->type, 0);

      ov = stacks_tmp->address;
      WRITEOUT_OV

      ov = stacks_tmp->linenumber;
      WRITEOUT_OV

      ov = stacks_tmp->bank;
      WRITEOUT_OV

      fprintf(final_ptr, "%c%c", stacks_tmp->slot, stacks_tmp->stacksize);

      for (ind = 0; ind < stacks_tmp->stacksize; ind++) {
	fprintf(final_ptr, "%c%c", stacks_tmp->stack[ind].type, stacks_tmp->stack[ind].sign);
	if (stacks_tmp->stack[ind].type == STACK_ITEM_TYPE_STRING)
	  fprintf(final_ptr, "%s%c", stacks_tmp->stack[ind].string, 0);
	else {
	  ov = stacks_tmp->stack[ind].value;
	  WRITEOUT_OV
	}
      }

      stacks_tmp = stacks_tmp->next;
    }

    /* data area */
    ind = 0;
    for (inz = 0; inz < max_address; inz++) {
      if (rom_banks_usage_table[inz] != 0) {
	/* data block id */
	fprintf(final_ptr, "%c", 0x0);
	for (i = inz, ind = 0; inz < max_address; inz++, ind++)
	  if (rom_banks_usage_table[inz] == 0) {

	    ov = i;
	    WRITEOUT_OV

	    ov = ind;
	    WRITEOUT_OV

	    fwrite(&rom_banks[i], 1, ind, final_ptr);

	    ind = 0;
	    break;
	  }
      }
    }

    if (ind != 0) {
      ov = i;
      WRITEOUT_OV

      ov = ind;
      WRITEOUT_OV

      fwrite(&rom_banks[i], 1, ind, final_ptr);
    }

    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      /* section block id */
      fprintf(final_ptr, "%c%s%c", 0x1, sec_tmp->name, sec_tmp->status);

      ov = sec_tmp->address;
      WRITEOUT_OV

      ov = sec_tmp->bank;
      WRITEOUT_OV

      fprintf(final_ptr, "%c", sec_tmp->slot);

      ov = sec_tmp->size;
      WRITEOUT_OV

      fwrite(sec_tmp->data, 1, sec_tmp->size, final_ptr);

      sec_tmp = sec_tmp->next;
    }

    fclose(final_ptr);
  }

  /* show project information */
  if (verbose_mode == ON && output_format != OUTPUT_LIBRARY) {
    x = 0;
    for (ind = 0; ind < max_address; ind++) {
      if (rom_banks_usage_table[ind] == 0 && x == 0) {
	x = 1;
	inz = ind;
      }
      else if (rom_banks_usage_table[ind] != 0 && x == 1) {
	if (inz == (ind - 1))
	  fprintf(stderr, "Free space at $%.4x.\n", inz);
	else
	  fprintf(stderr, "Free space at $%.4x-$%.4x.\n", inz, ind - 1);
	x = 0;
      }
    }

    if (x == 1) {
      if (inz == (ind - 1))
	fprintf(stderr, "Free space at $%.4x.\n", inz);
      else
	fprintf(stderr, "Free space at $%.4x-$%.4x.\n", inz, ind - 1);
    }

    for (ind = 0, q = 0; ind < max_address; q++) {
      for (x = 0, inz = 0; inz < banks[q]; inz++) {
	if (rom_banks_usage_table[ind++] == 0)
	  x++;
      }
      printf("Bank %.2d has %.5d bytes (%.2f%%) free.\n", q, x, (((double)x) / banks[q]) * 100);
    }

#ifndef GB
    if (output_format == OUTPUT_PROGRAM) {
      for (ind = binary_start, inz = 0; ind < binary_start + final_size; ind++) {
	if (rom_banks_usage_table[ind] == 0)
	  inz++;
      }
      fprintf(stderr, "%d unused bytes (%.2f%%) of total %d.\n", inz, (((double)inz) / final_size) * 100, final_size);
    }
    else {
#endif
      for (ind = 0, inz = 0; ind < max_address; ind++) {
	if (rom_banks_usage_table[ind] == 0)
	  inz++;
      }
      fprintf(stderr, "%d unused bytes of total %d.\n", inz, max_address);
#ifndef GB
    }
#endif

#ifndef GB
    if (header_defined != 0)
      fprintf(stderr, "File header size %d.\n", header_size);
    if (footer_defined != 0)
      fprintf(stderr, "File footer size %d.\n", footer_size);
    ind = header_size + footer_size;

    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      if (sec_tmp->status == SECTION_STATUS_HEADER) {
	fprintf(stderr, "Bank %d header section size %d.\n", sec_tmp->bank, sec_tmp->size);
	ind += sec_tmp->size;
      }
      sec_tmp = sec_tmp->next;
    }

    if (ind != 0) {
      fprintf(stderr, "Total %d additional bytes (from headers and footers).\n", ind);
      if (output_format == OUTPUT_PROGRAM)
	fprintf(stderr, "Total size %d bytes.\n", ind + final_size);
      else
	fprintf(stderr, "Total size %d bytes.\n", ind + max_address);
    }
#endif

  }
  else if (verbose_mode == ON && output_format == OUTPUT_LIBRARY) {

    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      printf("Section \"%s\" size %d.\n", sec_tmp->name, sec_tmp->size);
      sec_tmp = sec_tmp->next;
    }
  }

  return SUCCEEDED;
}


int mem_insert(unsigned char x) {

  if (section_status == ON) {
    sec_tmp->data[sec_tmp->i] = x;
    sec_tmp->i++;
    pc_bank++;
    pc_full++;
    pc_slot++;

    return SUCCEEDED;
  }

  if (pc_bank >= banksize) {
    fprintf(stderr, "MEM_INSERT: Origin (%d) overflows from bank (%d).\n", pc_bank, rom_bank);
    return FAILED;
  }
  else if (pc_full >= max_address) {
    fprintf(stderr, "MEM_INSERT: The current address ($%.4x) exceeds the size of the ROM ($%.4x).\n", pc_full, max_address);
    return FAILED;
  }
  else if (pc_slot >= pc_slot_max) {
    fprintf(stderr, "MEM_INSERT: The current address ($%.4x) overflows from SLOT %d.\n", pc_slot, slot);
    return FAILED;
  }

  if (rom_banks_usage_table[pc_full] != 0 && rom_banks[pc_full] != x && mem_insert_overwrite == OFF)
    fprintf(stderr, "MEM_INSERT: %d. write into $%.4x (old: $%.2x, new: $%.2x).\n", rom_banks_usage_table[pc_full], pc_full, rom_banks[pc_full], x & 0xFF);

  rom_banks_usage_table[pc_full] = 2;
  rom_banks[pc_full] = x;
  pc_bank++;
  pc_full++;
  pc_slot++;

  return SUCCEEDED;
}


int mem_insert_pad(void) {

  if (section_status == ON) {
    sec_tmp->i++;
    pc_bank++;
    pc_full++;
    pc_slot++;

    return SUCCEEDED;
  }

  if (pc_bank >= banksize) {
    fprintf(stderr, "MEM_INSERT_PAD: Origin (%d) overflows from bank (%d).\n", pc_bank, rom_bank);
    return FAILED;
  }
  else if (pc_full >= max_address) {
    fprintf(stderr, "MEM_INSERT_PAD: The current address ($%.4x) exceeds the size of the ROM ($%.4x).\n", pc_full, max_address);
    return FAILED;
  }
  else if (pc_slot >= pc_slot_max) {
    fprintf(stderr, "MEM_INSERT_PAD: The current address ($%.4x) overflows from SLOT %d.\n", pc_slot, slot);
    return FAILED;
  }

  /* announce the overwrite later */
  if (!(rom_banks_usage_table[pc_full] != 0 && mem_insert_overwrite == OFF))
    rom_banks_usage_table[pc_full] = 1;

  pc_bank++;
  pc_full++;
  pc_slot++;

  return SUCCEEDED;
}


int mem_insert_fix(unsigned char x) {

  if (section_status == ON) {
    sec_tmp->data[sec_tmp->i] = x;
    sec_tmp->i++;
    pc_bank++;
    pc_full++;
    pc_slot++;

    return SUCCEEDED;
  }

  if (pc_bank >= banksize) {
    fprintf(stderr, "MEM_INSERT: Origin (%d) overflows from bank (%d).\n", pc_bank, rom_bank);
    return FAILED;
  }
  else if (pc_full >= max_address) {
    fprintf(stderr, "MEM_INSERT: The current address ($%.4x) exceeds the size of the ROM ($%.4x).\n", pc_full, max_address);
    return FAILED;
  }
  else if (pc_slot >= pc_slot_max) {
    fprintf(stderr, "MEM_INSERT: The current address ($%.4x) overflows from SLOT %d.\n", pc_slot, slot);
    return FAILED;
  }

  if (rom_banks_usage_table[pc_full] > 1 && rom_banks[pc_full] != x && mem_insert_overwrite == OFF)
    fprintf(stderr, "MEM_INSERT: %d. write into $%.4x (old: $%.2x, new: $%.2x).\n", rom_banks_usage_table[pc_full], pc_full, rom_banks[pc_full], x & 0xFF);

  if (mem_insert_overwrite == ON)
    rom_banks_usage_table[pc_full]++;
  rom_banks[pc_full] = x;
  pc_bank++;
  pc_full++;
  pc_slot++;

  return SUCCEEDED;
}


int mem_insert_absolute(unsigned int add, unsigned char x) {

  if (add >= max_address) {
    fprintf(stderr, "MEM_INSERT_ABSOLUTE: The current address ($%.4x) exceeds the size of the ROM ($%.4x).\n", add, max_address);
    return FAILED;
  }

  if (rom_banks_usage_table[add] > 1 && rom_banks[add] != x && mem_insert_overwrite == OFF)
    fprintf(stderr, "MEM_INSERT_ABSOLUTE: %d. write into $%.4x (old: $%.2x, new: $%.2x).\n", rom_banks_usage_table[add], add, rom_banks[add], x & 0xFF);

  rom_banks_usage_table[add]++;
  rom_banks[add] = x;

  return SUCCEEDED;
}


int output_section(int type) {

  int i, o = 0, f;


  switch (type) {

  case SECTION_STATUS_ABSOLUTE:

    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      if (sec_tmp->status != SECTION_STATUS_ABSOLUTE) {
	sec_tmp = sec_tmp->next;
	continue;
      }

      inz = sec_tmp->address;

      for (i = 0; i < sec_tmp->size; i++)
	mem_insert_absolute(inz + i, sec_tmp->data[i]);

      sec_tmp = sec_tmp->next;
    }

    break;

  case SECTION_STATUS_FORCE:

    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      if (sec_tmp->status != SECTION_STATUS_FORCE) {
	sec_tmp = sec_tmp->next;
	continue;
      }

      inz = sec_tmp->address + bankaddress[sec_tmp->bank];
      i = inz + sec_tmp->size;
      f = SUCCEEDED;

      for (o = inz; inz < i; inz++)
	if (rom_banks_usage_table[inz] != 0) {
	  f = FAILED;
	  break;
	}

      if (f == SUCCEEDED) {
	pc_full = o;
	/* update the section starting address */
	sec_tmp->address = o - bankaddress[sec_tmp->bank];
	pc_bank = sec_tmp->address;
	if (banksize_defined == 0)
	  banksize = banks[sec_tmp->bank];

	/* get slot address */
	pc_slot = slots[sec_tmp->slot].address + pc_bank;
	pc_slot_max = slots[sec_tmp->slot].address + slots[sec_tmp->slot].size;

	for (i = 0; i < sec_tmp->size; i++)
	  mem_insert_fix(sec_tmp->data[i]);
      }
      else {
	fprintf(stderr, "INTERNAL_PASS_2: No room for section \"%s\".\n", sec_tmp->name);
	return FAILED;
      }

      sec_tmp = sec_tmp->next;
    }

    break;

  case SECTION_STATUS_FREE:

    sec_tmp = sections_first;
    while (sec_tmp != NULL) {
      if (!(sec_tmp->status == SECTION_STATUS_FREE || sec_tmp->status == SECTION_STATUS_SEMIFREE)) {
	sec_tmp = sec_tmp->next;
	continue;
      }

      f = FAILED;
      inz = bankaddress[sec_tmp->bank] + sec_tmp->address;
      i = bankaddress[sec_tmp->bank] + banks[sec_tmp->bank];

      for (; inz < i; inz++) {
	if (rom_banks_usage_table[inz] == 0) {
	  o = inz;
	  for (ind = 0; ind < sec_tmp->size; ind++, inz++) {
	    if (rom_banks_usage_table[inz] != 0)
	      break;
	  }
	  if (ind == sec_tmp->size) {
	    f = SUCCEEDED;
	    break;
	  }
	}
      }

      if (f == SUCCEEDED) {
	pc_full = o;
	/* update the section starting address */
	sec_tmp->address = o - bankaddress[sec_tmp->bank];
	pc_bank = sec_tmp->address;
	if (banksize_defined == 0)
	  banksize = banks[sec_tmp->bank];

	/* get slot address */
	pc_slot = slots[sec_tmp->slot].address + pc_bank;
	pc_slot_max = slots[sec_tmp->slot].address + slots[sec_tmp->slot].size;

	for (i = 0; i < sec_tmp->size; i++)
	  mem_insert_fix(sec_tmp->data[i]);
      }
      else {
	fprintf(stderr, "INTERNAL_PASS_2: No room for section \"%s\".\n", sec_tmp->name);
	return FAILED;
      }

      sec_tmp = sec_tmp->next;
    }

    break;

  case SECTION_STATUS_OVERWRITE:

    sec_tmp = sections_first;
    mem_insert_overwrite = ON;
    while (sec_tmp != NULL) {
      if (sec_tmp->status != SECTION_STATUS_OVERWRITE) {
	sec_tmp = sec_tmp->next;
	continue;
      }

      o = sec_tmp->address + bankaddress[sec_tmp->bank];
      pc_full = o;
      pc_bank = sec_tmp->address;
      if (banksize_defined == 0)
	banksize = banks[sec_tmp->bank];

      /* get slot address */
      pc_slot = slots[sec_tmp->slot].address + pc_bank;
      pc_slot_max = slots[sec_tmp->slot].address + slots[sec_tmp->slot].size;

      for (i = 0; i < sec_tmp->size; i++)
	mem_insert_fix(sec_tmp->data[i]);

      sec_tmp = sec_tmp->next;
    }

    mem_insert_overwrite = OFF;
    break;
  }

  return SUCCEEDED;
}


int export_definitions(FILE *final_ptr) {

  ov = 0;
  export_tmp = export_first;
  while (export_tmp != NULL) {
    tmp_def = defines;
    while (tmp_def != NULL) {
      if (strcmp(tmp_def->alias, export_tmp->name) == 0) {
	if (tmp_def->type == DEFINITION_TYPE_VALUE)
	  ov++;
	break;
      }
      tmp_def = tmp_def->next;
    }
    export_tmp = export_tmp->next;
  }

  WRITEOUT_OV

  export_tmp = export_first;
  while (export_tmp != NULL) {

    tmp_def = defines;
    while (tmp_def != NULL) {
      if (strcmp(tmp_def->alias, export_tmp->name) == 0)
	break;
      tmp_def = tmp_def->next;
    }

    if (tmp_def == NULL)
      fprintf(stderr, "WARNING: Trying to export an unkonwn definition (\"%s\").\n", export_tmp->name);
    else {
      if (tmp_def->type == DEFINITION_TYPE_VALUE) {
	fprintf(final_ptr, "%s%c", tmp_def->alias, 0x0);
	ov = tmp_def->value;
	WRITEOUT_OV
      }
      else
	fprintf(stderr, "INTERNAL_PASS_2: Definition \"%s\" is a string definition, which cannot be exported.\n", export_tmp->name);
    }

    export_tmp = export_tmp->next;
  }

  return SUCCEEDED;
}


#ifdef W65816
int get_snes_cpu_bank(struct label_def *l) {

  struct section_def *s;
  int x, k;


  if (l->section_status == OFF)
    k = bankaddress[l->bank] + l->address;
  else {
    s = sections_first;
    while (s != NULL) {
      if (l->section_id == s->id)
	break;
      else
	s = s->next;
    }
    k = s->address + l->address + bankaddress[s->bank];
  }

  if (lorom_defined == 1)
    x = (k / 0x8000) + l->base;
  else
    x = (k / 0x10000) + l->base;

  return x;
}
#endif
